/*
 * xsky - a computerized sky atlas
 *
 * Copyright 1992-3, Terry R. Friedrichsen
 *
 * This program may be copied and redistributed, in whole or in part,
 * as long as you don't try to make any money from the sale or redis-
 * tribution of the program or any part of the program, or pretend
 * that you wrote the program or any of its parts unless specifically
 * credited by the original author.
 *
 * You are free to make use of this software in your own programs, as
 * long as you credit the original author where it is due.
 */

/*
 * WARRANTY:
 * xsky was written as a learning project and as a demonstration of
 * X Window System programming.  xsky doesn't do anything; it is not
 * merchantable, and it is not fit for any purpose whatsoever.  In
 * fact, don't use xsky at all; it's free, and you're getting what
 * you paid for.
 */

#include <stdio.h>

#include <math.h>
#include <stdlib.h>

#include <unistd.h>

#include <X11/Intrinsic.h>

#include "rngc.h"

#include "skydefs.h"

#include "pos.h"
#include "catalog.h"
#include "draw.h"
#include "catreader.h"

struct rngc_node {
  double ra_rad;
  double dec_rad;
  long filepos;
  shortint rngc_number;
  char subletter;
  smallint objtype;
};

#ifdef __STDC__
/* static function prototypes */
static FILE *read_rngc();
static void calculate_rngc_field(struct display *,struct cat_header *);
static boolean find_rngc_obj(char *);
static void get_rngc_pos(int,double *,double *);
static void draw_rngc_obj(Widget,int,int,int,int,int,int,int);
static boolean ps_draw_rngc_obj(struct obj_node *,int);
static boolean display_rngc(FILE *,int);
static void format_rngc_name(int,char *,IDfont *);
static void fill_rngc_entry(char *,int);
static boolean find_messier_obj(char *);

/* external function prototypes */
extern char *build_filespec(char *,char *);
extern boolean read_binary_database(char *,void *,int *);
extern void init_precess(double,double);
extern boolean write_binary_database(char *,void *,size_t,size_t);
extern void read_desc_table();
extern double ra_pos_to_rad(struct ra_pos);
extern double dec_pos_to_rad(struct dec_pos);
extern void precess(double,double,double *,double *);
extern void add_to_display_list(struct display *,struct cat_header *,int,
				                               double,double);
extern void pos_rad_to_X(struct display *,double,double,int *,int *);
extern void draw_object(Widget,Position,Position,int,ObjShape);
extern boolean ps_draw_object(double,double,Dimension,double,ObjShape);
extern void format_rngc_display(char *);
extern void update_display_pos(double,double);
#else   /* not __STDC__ */
/* static function prototypes */
static FILE *read_rngc();
static void calculate_rngc_field();
static boolean find_rngc_obj();
static void get_rngc_pos();
static void draw_rngc_obj();
static boolean ps_draw_rngc_obj();
static boolean display_rngc();
static void format_rngc_name();
static void fill_rngc_entry();
static boolean find_messier_obj();

/* external function prototypes */
extern char *build_filespec();
extern boolean read_binary_database();
extern void init_precess();
extern boolean write_binary_database();
extern void read_desc_table();
extern double ra_pos_to_rad();
extern double dec_pos_to_rad();
extern void precess();
extern void add_to_display_list();
extern void pos_rad_to_X();
extern void draw_object();
extern boolean ps_draw_object();
extern void format_rngc_display();
extern void update_display_pos();
#endif   /* __STDC__ */

extern int messier_trans[];

static struct rngc_node rngc_array[RNGC_NUMREC];

static int rngc_array_max;

extern struct obj_node *freelist;

#define RNGC_DATA_FILE   "rngc.dat"
#define RNGC_BINARY_FILE "rngc.bin"

/* RNGC record buffer */
static char rngc_rec[RNGC_RECLEN + 1 + 1];



/* initialize the catalog data structure */

void init_rngc(catalog)

struct cat_header *catalog;

{
  catalog->read_data = read_rngc;
  catalog->calculate_field = calculate_rngc_field;
  catalog->mag_limit_check = NULLBOOLFUNCTION;
  catalog->find_object = find_rngc_obj;
  catalog->get_object_pos = get_rngc_pos;
  catalog->draw_object = draw_rngc_obj;
  catalog->ps_draw_object = ps_draw_rngc_obj;
  catalog->display_data = display_rngc;
  catalog->format_name = format_rngc_name;
  catalog->menu_name = RNGC_MENU_NAME;
  catalog->catalog_name = RNGC_CATALOG_NAME;

  return;
}






/* read in the RNGC catalog, extracting needed info */

static FILE *read_rngc()

{
  char *filebuff;
  FILE *rngc_fd;
  char *bin_filebuff;
  int rngc_entry;

/* get the path to the RNGC data */
  filebuff = build_filespec("RNGC_PATH",RNGC_DATA_FILE);

/* open the RNGC data file */
  if ((rngc_fd = fopen(filebuff,"r")) == (FILE *)NULL) {
    printf("sky:  can't open RNGC data file %s\n",filebuff);
    perror("sky");
    exit(1);
  }

/* get the path to the RNGC binary file */
  bin_filebuff = build_filespec("RNGC_BIN_PATH",RNGC_BINARY_FILE);

/* see if we can read the binary file */
  if (access(bin_filebuff,R_OK) == 0)
    /* we can; try to use the pre-built database */
    if (read_binary_database(bin_filebuff,rngc_array,&rngc_array_max)) {
      /* success; free the filename buffers */
      free(filebuff);
      free(bin_filebuff);

      /* read in the table of RNGC description abbreviations */
      read_desc_table();

      /* return the fd of the open data file */
      return(rngc_fd);
    }

/* prior to reading the RNGC catalog, build the rotation matrix for precessing
 * the object coordinates to the common equinox from the given epoch */
  init_precess(RNGC_EPOCH,COMMON_EQUINOX);

/* read the RNGC catalog, saving the information necessary for the display */
  rngc_entry = 0;
  while (TRUE) {
    rngc_array[rngc_entry].filepos = ftell(rngc_fd);
    if (fgets(rngc_rec,sizeof(rngc_rec),rngc_fd) == (char *)NULL)
      break;
    fill_rngc_entry(rngc_rec,rngc_entry++);
  }

/* leave the file open for later use */
/*  fclose(rngc_fd);*/

/* remember the number of RNGC entries we filled */
  rngc_array_max = rngc_entry - 1;

/* attempt to save the binary database */
  if (access(bin_filebuff,W_OK) == 0)
    /* it's writeable, so let's try */
    if (! write_binary_database(bin_filebuff,rngc_array,
				           (size_t)sizeof(struct rngc_node),
				                          (size_t)rngc_entry))
      /* dump of binary database failed; be sure the file is gone */
      (void)unlink(bin_filebuff);

/* free the filename buffers */
  free(filebuff);
  free(bin_filebuff);

/* read in the table of RNGC description abbreviations */
  read_desc_table();

  return(rngc_fd);
}



/* given an object's index, return its right ascension and declination */

static void get_rngc_pos(array_pos,ra_rad,dec_rad)

int array_pos;
double *ra_rad, *dec_rad;

{
  *ra_rad = rngc_array[array_pos].ra_rad;
  *dec_rad = rngc_array[array_pos].dec_rad;

  return;
}



/* given an RNGC record, make a data node for it */

static void fill_rngc_entry(record,rngc_entry)

char *record;
int rngc_entry;

{
  int reclen;
  char field[11];
  struct ra_pos ra_pos;
  struct dec_pos dec_pos;

/* get the length of the record for GET_FIELD */
  reclen = strlen(record) - 1;

/* get the RNGC number */
  GET_FIELD(field,record,CAT_START,5);
  rngc_array[rngc_entry].rngc_number = atoi(field);

/* get the RNGC sub-letter */
  GET_FIELD(field,record,SUBLETTER_START,1);
  rngc_array[rngc_entry].subletter = field[0];

/* get the object type */
  GET_FIELD(field,record,TYPE_START,2);
  rngc_array[rngc_entry].objtype = atoi(field);

/* get the RA in pieces */
  GET_FIELD(field,record,RA_START,2);
  ra_pos.hours = (smallint)atoi(field);

  GET_FIELD(field,record,RA_START + 2,2);
  ra_pos.minutes = (smallint)atoi(field);

  GET_FIELD(field,record,RA_START + 5,1);
  ra_pos.seconds = (smallint)(atoi(field) * (60 / 10));

  ra_pos.thousandths = (shortint)0;

/* similarly, get the declination in pieces */
  GET_FIELD(field,record,DEC_START,1);
  dec_pos.sign = (short)1;
  if (field[0] == '-')
    dec_pos.sign = (short)-1;

  GET_FIELD(field,record,DEC_START + 1,2);
  dec_pos.degrees = (smallint)atoi(field);

  GET_FIELD(field,record,DEC_START + 3,2);
  dec_pos.minutes = (smallint)atoi(field);

  dec_pos.seconds = (smallint)0;
  dec_pos.hundredths = (smallint)0;

/* compute the RA and declination in radians */
  rngc_array[rngc_entry].ra_rad = ra_pos_to_rad(ra_pos);
  rngc_array[rngc_entry].dec_rad = dec_pos_to_rad(dec_pos);

/* precess them to the common equinox */
  precess(rngc_array[rngc_entry].ra_rad,rngc_array[rngc_entry].dec_rad,
	       &rngc_array[rngc_entry].ra_rad,&rngc_array[rngc_entry].dec_rad);

  return;
}



/* construct a list of those objects which appear in the current display */

static void calculate_rngc_field(display,catalog)

struct display *display;
struct cat_header *catalog;

{
  int i;
  int objtype;

/* loop through all RNGC entries, checking for visibility */
  for (i = 0; i <= rngc_array_max; i++) {
    /* if the object is nonexistent, just ignore it */
    objtype = rngc_array[i].objtype;
    if (objtype >= 10)
      objtype /= 10;
    if (objtype == NONEXISTENT)
      continue;

    /* check this object and add it to the list if visible */
    add_to_display_list(display,catalog,i,rngc_array[i].ra_rad,
                                                       rngc_array[i].dec_rad);
  }

#ifdef DEBUG
  printf("calculate_rngc_field returning\n");
#endif
  return;
}



/* draw an RNGC object on the sky */

static void draw_rngc_obj(sky_widget,array_pos,obj_x,obj_y,ul_x,ul_y,
		                                                 width,height)

Widget sky_widget;
int array_pos;
Position obj_x, obj_y;
int ul_x, ul_y;
int width, height;

{
  int objtype;

/* get the major object type */
  objtype = rngc_array[array_pos].objtype;
  if (objtype >= 10)
    objtype /= 10;

#ifdef DEBUG
  printf("draw_obj called to draw figure type %d at %d,%d\n",objtype,
	                                                         obj_x,obj_y);
#endif

/* if this is a non-existent object, suppress the display */
  if (objtype == NONEXISTENT)
    return;

 /* draw the object even if it is only partially in the box */
  if (((obj_x + (OBJSIZE + 1) / 2) >= ul_x) &&
                            ((obj_x - (OBJSIZE + 1) / 2) <= (ul_x + width)) &&
                            ((obj_y + (OBJSIZE + 1) / 2) >= ul_y) &&
                            ((obj_y - (OBJSIZE + 1) / 2) <= (ul_y + height)))
    ;
  else
    return;

  if (objtype == GALAXY)
    /* draw an ellipse for a galaxy */
    draw_object(sky_widget,obj_x,obj_y,OBJSIZE,ELLIPSE);
  else if ((objtype == OPEN_CLUSTER) || (objtype == GLOBULAR_CLUSTER))
    /* draw a circle for open and globular clusters */
    draw_object(sky_widget,obj_x,obj_y,OBJSIZE,CIRCLE);
  else
    /* for the moment, draw a square for anything else */
    draw_object(sky_widget,obj_x,obj_y,OBJSIZE,SQUARE);

  return;
}



/* draw one RNGC object on the PostScript display */

static boolean ps_draw_rngc_obj(obj,height)

struct obj_node *obj;
Dimension height;

{
  int objtype;
  ObjShape shape;

/* get the major object type */
  objtype = rngc_array[obj->array_pos].objtype;
  if (objtype >= 10)
    objtype /= 10;

/* if this is a non-existent object, suppress the display */
  if (objtype == NONEXISTENT)
    return(TRUE);
  else if (objtype == GALAXY)
    /* draw an ellipse for a galaxy */
    shape = ELLIPSE;
  else if ((objtype == OPEN_CLUSTER) || (objtype == GLOBULAR_CLUSTER))
    /* draw a circle for open and globular clusters */
    shape = CIRCLE;
  else
    /* for the moment, draw a square for anything else */
    shape = SQUARE;

/* draw the object in the correct position with the correct size */
  return(ps_draw_object((double)obj->x,(double)obj->y,height,(double)OBJSIZE,
			                                              shape));
}



/* display the catalog data for the given object */

static boolean display_rngc(rngc_fd,rngc_array_pos)

FILE *rngc_fd;
int rngc_array_pos;

{
  int i;

/* get info from the RNGC catalog */
  if (fseek(rngc_fd,rngc_array[rngc_array_pos].filepos,SEEK_SET) == -1)
      return(FALSE);
  if (fgets(rngc_rec,RNGC_RECLEN + 2,rngc_fd) == (char *)NULL)
      return(FALSE);

/* fill out the (possibly short) record with blanks */
  for (i = strlen(rngc_rec) - 1; i < RNGC_RECLEN; i++)
    rngc_rec[i] = ' ';

/* format the record into an informational display */
  format_rngc_display(rngc_rec);

  return(TRUE);
}



/* find the RNGC object in the buffer and update the display coordinates */

static boolean find_rngc_obj(buffer)

char *buffer;

{
  int rngc_no;
  char rngc_subletter;
  int x, l, r;

/* check to see if this is an RNGC designation and step to the number */
  if ((strncmp(buffer,"NGC",3) == EQUAL) || strncmp(buffer,"ngc",3) == EQUAL)
    buffer = &buffer[3];
  else if ((strncmp(buffer,"RNGC",4) == EQUAL) ||
	                                   strncmp(buffer,"rngc",4) == EQUAL)
    buffer = &buffer[4];
  else
    /* if the first letter is 'M", try a Messier object */
    if ((*buffer == 'M') || (*buffer == 'm'))
	return(find_messier_obj(&buffer[1]));
    else
	return(FALSE);

  /* skip white space */
  while (isspace(*buffer))
    buffer++;

/* get the RNGC number */
  rngc_no = atoi(buffer);

/* find the letter suffix, if any */

  /* skip decimal digits */
  while (isdigit(*buffer))
    buffer++;

  /* skip more white space */
  while (isspace(*buffer))
    buffer++;

  /* grab the letter */
  if (isalpha(*buffer))
    rngc_subletter = toupper(*buffer);
  else
    rngc_subletter = ' ';

/* binary-search looking for the number (see Sedgewick's "Algorithms") */
  l = 1;
  r = rngc_array_max + 1;

  do {
    x = (l + r) / 2;
    if (rngc_no < rngc_array[x - 1].rngc_number)
      r = x - 1;
    else
      l = x + 1;
  } while (! ((rngc_no == rngc_array[x - 1].rngc_number) || (l > r)));

  /* return FALSE if we didn't find the RNGC object */
  if (rngc_no == rngc_array[x - 1].rngc_number)
    ;
  else
    return(FALSE);

/* adjust x to zero-relative index */
  x--;

/* back up to the beginning of this RNGC number */
  while (x > 0)
    if (rngc_array[x - 1].rngc_number == rngc_no)
      x--;
    else
      break;

/* search for the RNGC number with the matching subletter */
  while ((x <= rngc_array_max) && (rngc_array[x].rngc_number == rngc_no))
    if (rngc_subletter == rngc_array[x].subletter)
      break;
    else
      x++;

/* if we have a match, update the display center */
  if ((rngc_no == rngc_array[x].rngc_number) &&
                                (rngc_subletter == rngc_array[x].subletter)) {
    /* update the display coordinates */
    update_display_pos(rngc_array[x].ra_rad,rngc_array[x].dec_rad);
    return(TRUE);
  }
  else
    return(FALSE);

/* NOTREACHED */
}



/* find the Messier object in the buffer and update the display coordinates */

static boolean find_messier_obj(buffer)

char *buffer;

{
  char *ptr;
  int messier_number;
  char rngc_buffer[10];

/* skip white space */
  while (isspace(*buffer))
    buffer++;

/* is the rest of the buffer numeric?  if so, try it as a Messier number */
  ptr = buffer;
  while (isdigit(*ptr))
    ptr++;
  if (*ptr != '\0')
    return(FALSE);

/* get the Messier number */
  messier_number = atoi(buffer);

/* ensure it's reasonableness */
  if ((messier_number < 1) || (messier_number > NUM_MESSIER_OBJS))
    return(FALSE);
  if (messier_trans[messier_number - 1] == 0)
    return(FALSE);

/* build an RNGC buffer */
  sprintf(rngc_buffer,"RNGC %d",messier_trans[messier_number - 1]);

/* pretend the user entered the RNGC number */
  return(find_rngc_obj(rngc_buffer));
}



/* format a nice name for the RNGC object */

/* ARGSUSED */

static void format_rngc_name(rngc_pos,buffer,font)

int rngc_pos;
char *buffer;
IDfont *font;

{
  int i;

/* see if this is a Messier object */
  for (i = 0; i < NUM_MESSIER_OBJS; i++)
    if (messier_trans[i] == rngc_array[rngc_pos].rngc_number)
      break;

/* if we fell out of the loop, this is not a Messier object */
  if ((i >= NUM_MESSIER_OBJS) || (messier_trans[i] == 0))
    sprintf(buffer,"NGC %d",rngc_array[rngc_pos].rngc_number);
  else
    /* otherwise, use the Messier designation */
    sprintf(buffer,"M %d",i + 1);

  return;
}
